@@ -25,7 +25,7 @@ Dialog::Dialog(const QString& title, QWidget *parent,
                UseStatusLine useStatusLine, const QString& applyTitle,
                QDialogButtonBox::StandardButtons addButtons)
     : Dialog(title
-        , QDialogButtonBox::Ok | QDialogButtonBox::Cancel | addButtons
+        , QDialogButtonBox::Ok | /*QDialogButtonBox::Cancel |*/ addButtons
         , parent, useStatusLine)
 {
     if (!applyTitle.isEmpty())
@@ -44,10 +44,6 @@ Dialog::Dialog(const QString& title, QDialogButtonBox::StandardButtons setButton
 {
     setWindowTitle(title);
 
-#if (QT_VERSION < QT_VERSION_CHECK(5, 10, 0))
-    setWindowFlags(windowFlags() & ~Qt::WindowContextHelpButtonHint);
-#endif
-
     connect(buttons, &QDialogButtonBox::clicked, this, &Dialog::buttonClicked);
 
     outerLayout.addWidget(buttons);
@@ -55,16 +51,21 @@ Dialog::Dialog(const QString& title, QDialogButtonBox::StandardButtons setButton
         outerLayout.addWidget(statusLabel);
 }
 
-void Dialog::addLayout(QLayout* l)
+void Dialog::addLayout(QLayout* l, int stretch)
 {
     int offset = 1 + (statusLabel != nullptr);
-    outerLayout.insertLayout(outerLayout.count() - offset, l);
+    outerLayout.insertLayout(outerLayout.count() - offset, l, stretch);
 }
 
-void Dialog::addWidget(QWidget* w)
+void Dialog::addWidget(QWidget* w, int stretch, Qt::Alignment alignment)
 {
     int offset = 1 + (statusLabel != nullptr);
-    outerLayout.insertWidget(outerLayout.count() - offset, w);
+    outerLayout.insertWidget(outerLayout.count() - offset, w, stretch, alignment);
+}
+
+QPushButton*Dialog::button(QDialogButtonBox::StandardButton which)
+{
+    return buttonBox()->button(which);
 }
 
 void Dialog::reactivate()
@@ -95,15 +96,20 @@ void Dialog::buttonClicked(QAbstractButton* button)
     switch (buttons->buttonRole(button))
     {
         case QDialogButtonBox::AcceptRole:
-            if (statusLabel)
-                statusLabel->setText(pendingApplyMessage);
-            setDisabled(true);
-            apply();
+        case QDialogButtonBox::YesRole:
+            if (validate())
+            {
+                if (statusLabel)
+                    statusLabel->setText(pendingApplyMessage);
+                setDisabled(true);
+                apply();
+            }
             break;
         case QDialogButtonBox::ResetRole:
             load();
             break;
         case QDialogButtonBox::RejectRole:
+        case QDialogButtonBox::NoRole:
             reject();
             break;
         default:
