@@ -24,7 +24,8 @@
 class QLineEdit;
 class QCheckBox;
 
-namespace QMatrixClient {
+namespace Quotient {
+    class AccountSettings;
     class Connection;
 }
 
@@ -32,23 +33,31 @@ class LoginDialog : public Dialog
 {
         Q_OBJECT
     public:
-        explicit LoginDialog(QWidget* parent = nullptr);
+        explicit LoginDialog(const QString& statusMessage,
+                             QWidget* parent = nullptr,
+                             const QStringList& knownAccounts = {});
+        explicit LoginDialog(const QString& statusMessage, QWidget* parent,
+                             const Quotient::AccountSettings& reloginData);
+        void setup(const QString &statusMessage);
         ~LoginDialog() override;
 
-        QMatrixClient::Connection* releaseConnection();
+        Quotient::Connection* releaseConnection();
         QString deviceName() const;
         bool keepLoggedIn() const;
 
     private slots:
         void apply() override;
+        void loginWithBestFlow();
+        void loginWithPassword();
+        void loginWithSso();
         
     private:
-        QLineEdit* serverEdit;
         QLineEdit* userEdit;
         QLineEdit* passwordEdit;
         QLineEdit* initialDeviceName;
+        QLineEdit* serverEdit;
         QCheckBox* saveTokenCheck;
 
-        QScopedPointer<QMatrixClient::Connection, QScopedPointerDeleteLater>
+        QScopedPointer<Quotient::Connection, QScopedPointerDeleteLater>
             m_connection;
 };
