@@ -23,103 +23,99 @@
 #include <QtQml> // for qmlRegisterType()
 
 #include "../quaternionroom.h"
+#include "../htmlfilter.h"
 #include <connection.h>
 #include <user.h>
 #include <settings.h>
+#include <events/encryptionevent.h>
 #include <events/roommemberevent.h>
 #include <events/simplestateevents.h>
 #include <events/redactionevent.h>
 #include <events/roomavatarevent.h>
-
-enum EventRoles {
-    EventTypeRole = Qt::UserRole + 1,
-    EventIdRole,
-    TimeRole,
-    SectionRole,
-    AboveSectionRole,
-    AuthorRole,
-    AboveAuthorRole,
-    ContentRole,
-    ContentTypeRole,
-    HighlightRole,
-    ReadMarkerRole,
-    SpecialMarksRole,
-    LongOperationRole,
-    AnnotationRole,
-    // For debugging
-    EventResolvedTypeRole,
-};
+#include <events/roomcreateevent.h>
+#include <events/roomtombstoneevent.h>
+#include <events/roomcanonicalaliasevent.h>
+#include <events/reactionevent.h>
 
 QHash<int, QByteArray> MessageEventModel::roleNames() const
 {
-    QHash<int, QByteArray> roles = QAbstractItemModel::roleNames();
-    roles[EventTypeRole] = "eventType";
-    roles[EventIdRole] = "eventId";
-    roles[TimeRole] = "time";
-    roles[SectionRole] = "section";
-    roles[AboveSectionRole] = "aboveSection";
-    roles[AuthorRole] = "author";
-    roles[AboveAuthorRole] = "aboveAuthor";
-    roles[ContentRole] = "content";
-    roles[ContentTypeRole] = "contentType";
-    roles[HighlightRole] = "highlight";
-    roles[ReadMarkerRole] = "readMarker";
-    roles[SpecialMarksRole] = "marks";
-    roles[LongOperationRole] = "progressInfo";
-    roles[AnnotationRole] = "annotation";
-    roles[EventResolvedTypeRole] = "eventResolvedType";
+    static const auto roles = [this] {
+        auto roles = QAbstractItemModel::roleNames();
+        roles.insert(EventTypeRole, "eventType");
+        roles.insert(EventIdRole, "eventId");
+        roles.insert(TimeRole, "time");
+        roles.insert(SectionRole, "section");
+        roles.insert(AboveSectionRole, "aboveSection");
+        roles.insert(AuthorRole, "author");
+        roles.insert(AboveAuthorRole, "aboveAuthor");
+        roles.insert(ContentRole, "content");
+        roles.insert(ContentTypeRole, "contentType");
+        roles.insert(HighlightRole, "highlight");
+        roles.insert(SpecialMarksRole, "marks");
+        roles.insert(LongOperationRole, "progressInfo");
+        roles.insert(AnnotationRole, "annotation");
+        roles.insert(EventResolvedTypeRole, "eventResolvedType");
+        roles.insert(RefRole, "refId");
+        roles.insert(ReactionsRole, "reactions");
+        return roles;
+    }();
     return roles;
 }
 
 MessageEventModel::MessageEventModel(QObject* parent)
     : QAbstractListModel(parent)
-    , m_currentRoom(nullptr)
 {
-    using namespace QMatrixClient;
+    using namespace Quotient;
+#if (QT_VERSION >= QT_VERSION_CHECK(5, 14, 0))
+    qmlRegisterAnonymousType<FileTransferInfo>("Quotient", 1);
+#else
     qmlRegisterType<FileTransferInfo>(); qRegisterMetaType<FileTransferInfo>();
-    qmlRegisterUncreatableType<EventStatus>("QMatrixClient", 1, 0, "EventStatus",
-        "EventStatus is not an creatable type");
+#endif
+    qmlRegisterUncreatableType<EventStatus>("Quotient", 1, 0, "EventStatus",
+        "EventStatus is not a creatable type");
+    // This could be a single line in changeRoom() but then there's a race
+    // condition between the model reset completion and the room property
+    // update in QML - connecting the two signals early on overtakes any QML
+    // connection to modelReset. Ideally the room property could use modelReset
+    // for its NOTIFY signal - unfortunately, moc doesn't support using
+    // parent's signals with parameters in NOTIFY
+    connect(this, &MessageEventModel::modelReset, //
+            this, &MessageEventModel::roomChanged);
 }
 
+QuaternionRoom* MessageEventModel::room() const { return m_currentRoom; }
+
 void MessageEventModel::changeRoom(QuaternionRoom* room)
 {
     if (room == m_currentRoom)
         return;
 
+    if (m_currentRoom)
+        qDebug() << "Disconnecting event model from"
+                 << m_currentRoom->objectName();
     beginResetModel();
-    if( m_currentRoom )
-    {
-        m_currentRoom->disconnect( this );
-        qDebug() << "Disconnected from" << m_currentRoom->id();
-    }
+    if (m_currentRoom)
+        m_currentRoom->disconnect(this);
 
     m_currentRoom = room;
-    if( room )
-    {
-        lastReadEventId = room->readMarkerEventId();
-
-        using namespace QMatrixClient;
+    if (m_currentRoom) {
+        using namespace Quotient;
         connect(m_currentRoom, &Room::aboutToAddNewMessages, this,
-                [=](RoomEventsRange events)
-                {
-                    beginInsertRows({}, timelineBaseIndex(),
-                                    timelineBaseIndex() + int(events.size()) - 1);
+                [this](RoomEventsRange events) {
+                    incomingEvents(events, timelineBaseIndex());
                 });
         connect(m_currentRoom, &Room::aboutToAddHistoricalMessages, this,
-                [=](RoomEventsRange events)
-                {
-                    if (rowCount() > 0)
-                        rowBelowInserted = rowCount() - 1; // See #312
-                    beginInsertRows({}, rowCount(),
-                                    rowCount() + int(events.size()) - 1);
+                [this](RoomEventsRange events) {
+                    incomingEvents(events, rowCount());
                 });
         connect(m_currentRoom, &Room::addedMessages, this,
-                [=] (int lowest, int biggest) {
+                [this] (int lowest, int biggest) {
                     endInsertRows();
                     if (biggest < m_currentRoom->maxTimelineIndex())
                     {
-                        auto rowBelowInserted =
-                                m_currentRoom->maxTimelineIndex() - biggest + timelineBaseIndex() - 1;
+                        const auto rowBelowInserted =
+                            m_currentRoom->maxTimelineIndex() - biggest
+                            + timelineBaseIndex() - 1;
                         refreshEventRoles(rowBelowInserted,
                                          {AboveAuthorRole, AboveSectionRole});
                     }
@@ -141,8 +137,9 @@ void MessageEventModel::changeRoom(QuaternionRoom* room)
                     movingEvent = true;
                     // Reverse i because row 0 is bottommost in the model
                     const auto row = timelineBaseIndex() - i - 1;
-                    Q_ASSERT(beginMoveRows({}, row, row,
-                                           {}, timelineBaseIndex()));
+                    auto moveBegan = beginMoveRows({}, row, row,
+                                                   {}, timelineBaseIndex());
+                    Q_ASSERT(moveBegan);
                 });
         connect(m_currentRoom, &Room::pendingEventMerged, this,
                 [this] {
@@ -153,9 +150,6 @@ void MessageEventModel::changeRoom(QuaternionRoom* room)
                     }
                     refreshRow(timelineBaseIndex()); // Refresh the looks
                     refreshLastUserEvents(0);
-                    if (m_currentRoom->timelineSize() > 1) // Refresh above
-                        refreshEventRoles(timelineBaseIndex() + 1,
-                                          {ReadMarkerRole});
                     if (timelineBaseIndex() > 0) // Refresh below, see #312
                         refreshEventRoles(timelineBaseIndex() - 1,
                                           {AboveAuthorRole, AboveSectionRole});
@@ -167,18 +161,14 @@ void MessageEventModel::changeRoom(QuaternionRoom* room)
         connect(m_currentRoom, &Room::pendingEventDiscarded,
                 this, &MessageEventModel::endRemoveRows);
         connect(m_currentRoom, &Room::readMarkerMoved,
-            this, [this] {
-            refreshEventRoles(
-                std::exchange(lastReadEventId,
-                              m_currentRoom->readMarkerEventId()),
-                {ReadMarkerRole});
-            refreshEventRoles(lastReadEventId, {ReadMarkerRole});
-        });
+                this, &MessageEventModel::readMarkerUpdated);
         connect(m_currentRoom, &Room::replacedEvent, this,
                 [this] (const RoomEvent* newEvent) {
                     refreshLastUserEvents(
                         refreshEvent(newEvent->id()) - timelineBaseIndex());
                 });
+        connect(m_currentRoom, &Room::updatedEvent,
+                this, &MessageEventModel::refreshEvent);
         connect(m_currentRoom, &Room::fileTransferProgress,
                 this, &MessageEventModel::refreshEvent);
         connect(m_currentRoom, &Room::fileTransferCompleted,
@@ -187,16 +177,21 @@ void MessageEventModel::changeRoom(QuaternionRoom* room)
                 this, &MessageEventModel::refreshEvent);
         connect(m_currentRoom, &Room::fileTransferCancelled,
                 this, &MessageEventModel::refreshEvent);
-        qDebug() << "Connected to room" << room->id()
+        qDebug() << "Event model connected to room" << room->objectName()
                  << "as" << room->localUser()->id();
-    } else
-        lastReadEventId.clear();
+    }
     endResetModel();
+    emit readMarkerUpdated();
 }
 
 int MessageEventModel::refreshEvent(const QString& eventId)
 {
-    return refreshEventRoles(eventId);
+    int row = findRow(eventId, true);
+    if (row >= 0)
+        refreshEventRoles(row);
+    else
+        qWarning() << "Trying to refresh inexistent event:" << eventId;
+    return row;
 }
 
 void MessageEventModel::refreshRow(int row)
@@ -204,6 +199,29 @@ void MessageEventModel::refreshRow(int row)
     refreshEventRoles(row);
 }
 
+void MessageEventModel::incomingEvents(Quotient::RoomEventsRange events,
+                                       int atIndex)
+{
+    beginInsertRows({}, atIndex, atIndex + int(events.size()) - 1);
+}
+
+int MessageEventModel::readMarkerVisualIndex() const
+{
+    if (!m_currentRoom)
+        return -1; // Beyond the bottommost (sync) edge of the timeline
+    if (auto r = findRow(m_currentRoom->readMarkerEventId()); r != -1) {
+        // Ensure that the read marker is on a visible event
+        // TODO: move this to libQuotient once it allows to customise
+        //       event status calculation
+        while (r < rowCount() - 1
+               && data(index(r, 0), SpecialMarksRole)
+                  == Quotient::EventStatus::Hidden)
+            ++r;
+        return r;
+    }
+    return rowCount(); // Beyond the topmost (history) edge of the timeline
+}
+
 int MessageEventModel::timelineBaseIndex() const
 {
     return m_currentRoom ? int(m_currentRoom->pendingEvents().size()) : 0;
@@ -215,55 +233,57 @@ void MessageEventModel::refreshEventRoles(int row, const QVector<int>& roles)
     emit dataChanged(idx, idx, roles);
 }
 
-int MessageEventModel::refreshEventRoles(const QString& eventId,
-                                         const QVector<int>& roles)
+int MessageEventModel::findRow(const QString& id, bool includePending) const
 {
-    const auto it = m_currentRoom->findInTimeline(eventId);
-    if (it == m_currentRoom->timelineEdge())
-    {
-        qWarning() << "Trying to refresh inexistent event:" << eventId;
-        return -1;
+    // On 64-bit platforms, difference_type for std containers is long long
+    // but Qt uses int throughout its interfaces; hence casting to int below.
+    if (!id.isEmpty()) {
+        // First try pendingEvents because it is almost always very short.
+        if (includePending) {
+            const auto pendingIt = m_currentRoom->findPendingEvent(id);
+            if (pendingIt != m_currentRoom->pendingEvents().end())
+                return int(pendingIt - m_currentRoom->pendingEvents().begin());
+        }
+        const auto timelineIt = m_currentRoom->findInTimeline(id);
+        if (timelineIt != m_currentRoom->historyEdge())
+            return int(timelineIt - m_currentRoom->messageEvents().rbegin())
+                    + timelineBaseIndex();
     }
-    const auto row =
-            it - m_currentRoom->messageEvents().rbegin() + timelineBaseIndex();
-    refreshEventRoles(row, roles);
-    return row;
+    return -1;
 }
 
-inline bool hasValidTimestamp(const QMatrixClient::TimelineItem& ti)
+inline bool hasValidTimestamp(const Quotient::TimelineItem& ti)
 {
-    return ti->timestamp().isValid();
+    return ti->originTimestamp().isValid();
 }
 
 QDateTime MessageEventModel::makeMessageTimestamp(
             const QuaternionRoom::rev_iter_t& baseIt) const
 {
     const auto& timeline = m_currentRoom->messageEvents();
-    auto ts = baseIt->event()->timestamp();
-    if (ts.isValid())
+    if (auto ts = baseIt->event()->originTimestamp(); ts.isValid())
         return ts;
 
     // The event is most likely redacted or just invalid.
     // Look for the nearest date around and slap zero time to it.
-    using QMatrixClient::TimelineItem;
-    auto rit = std::find_if(baseIt, timeline.rend(),
-                      hasValidTimestamp);
-    if (rit != timeline.rend())
-        return { rit->event()->timestamp().date(), {0,0}, Qt::LocalTime };
-    auto it = std::find_if(baseIt.base(), timeline.end(), hasValidTimestamp);
-    if (it != timeline.end())
-        return { it->event()->timestamp().date(), {0,0}, Qt::LocalTime };
+    if (auto rit = std::find_if(baseIt, timeline.rend(), hasValidTimestamp);
+            rit != timeline.rend())
+        return { rit->event()->originTimestamp().date(), {0,0}, Qt::LocalTime };
+    if (auto it = std::find_if(baseIt.base(), timeline.end(), hasValidTimestamp);
+            it != timeline.end())
+        return { it->event()->originTimestamp().date(), {0,0}, Qt::LocalTime };
 
     // What kind of room is that?..
     qCritical() << "No valid timestamps in the room timeline!";
     return {};
 }
 
-QString MessageEventModel::renderDate(const QDateTime& timestamp) const
+QString MessageEventModel::renderDate(const QDateTime& timestamp)
 {
     auto date = timestamp.toLocalTime().date();
-    if (QMatrixClient::SettingsGroup("UI")
-            .value("banner_human_friendly_date", true).toBool())
+    static Quotient::SettingsGroup sg { "UI" };
+    if (sg.get("use_human_friendly_dates",
+               sg.get("banner_human_friendly_date", true)))
     {
         if (date == QDate::currentDate())
             return tr("Today");
@@ -273,14 +293,15 @@ QString MessageEventModel::renderDate(const QDateTime& timestamp) const
             return tr("The day before yesterday");
         if (date > QDate::currentDate().addDays(-7))
         {
-            // Make sure to capitalise the day name.
             auto s = QLocale().standaloneDayName(date.dayOfWeek());
-            if (!s.isEmpty())
-                s[0] = QLocale().toUpper(s.mid(0,1))[0];
+            // Some locales (e.g., Russian on Windows) don't capitalise
+            // the day name so make sure the first letter is uppercase.
+            if (!s.isEmpty() && !s[0].isUpper())
+                s[0] = QLocale().toUpper(s.mid(0,1)).at(0);
             return s;
         }
     }
-    return date.toString(Qt::DefaultLocaleShortDate);
+    return QLocale().toString(date, QLocale::ShortFormat);
 }
 
 bool MessageEventModel::isUserActivityNotable(
@@ -296,12 +317,12 @@ bool MessageEventModel::isUserActivityNotable(
     // double-check that there are no redactions and that it's not a single
     // join or leave.
 
-    using namespace QMatrixClient;
+    using namespace Quotient;
     bool joinFound = false, redactionsFound = false;
     // Find the nearest join of this user above, or a no-nonsense event.
     for (auto it = baseIt,
               limit = baseIt +
-                    std::min(int(m_currentRoom->timelineEdge() - baseIt), 100);
+                    std::min(int(m_currentRoom->historyEdge() - baseIt), 100);
          it != limit; ++it)
     {
         const auto& e = **it;
@@ -390,16 +411,14 @@ int MessageEventModel::rowCount(const QModelIndex& parent) const
 {
     if( !m_currentRoom || parent.isValid() )
         return 0;
-    return m_currentRoom->timelineSize();
+    return m_currentRoom->timelineSize() + m_currentRoom->pendingEvents().size();
 }
 
 QVariant MessageEventModel::data(const QModelIndex& idx, int role) const
 {
     const auto row = idx.row();
 
-    if( !m_currentRoom || row < 0 ||
-            row >= int(m_currentRoom->pendingEvents().size()) +
-                       m_currentRoom->timelineSize())
+    if (!idx.isValid() || row >= rowCount())
         return {};
 
     bool isPending = row < timelineBaseIndex();
@@ -409,7 +428,7 @@ QVariant MessageEventModel::data(const QModelIndex& idx, int role) const
                                 std::min(row, timelineBaseIndex());
     const auto& evt = isPending ? **pendingIt : **timelineIt;
 
-    using namespace QMatrixClient;
+    using namespace Quotient;
     if( role == Qt::DisplayRole )
     {
         if (evt.isRedacted())
@@ -418,59 +437,88 @@ QVariant MessageEventModel::data(const QModelIndex& idx, int role) const
             if (reason.isEmpty())
                 return tr("Redacted");
 
-            return tr("Redacted: %1")
-                .arg(evt.redactedBecause()->reason());
+            return tr("Redacted: %1").arg(reason.toHtmlEscaped());
         }
 
+        // clang-format off
         return visit(evt
             , [this] (const RoomMessageEvent& e) {
+                // clang-format on
                 using namespace MessageEventContent;
 
-                if (e.hasTextContent() && e.mimeType().name() != "text/plain")
-                    return static_cast<const TextContent*>(e.content())->body;
-                if (e.hasFileContent())
-                {
-                    auto fileCaption = e.content()->fileInfo()->originalName;
+                if (e.hasTextContent() && e.mimeType().name() != "text/plain") {
+                    // Naïvely assume that it's HTML
+                    auto htmlBody =
+                        static_cast<const TextContent*>(e.content())->body;
+                    auto [cleanHtml, errorPos, errorString] =
+                        HtmlFilter::fromMatrixHtml(htmlBody, m_currentRoom);
+                    // If HTML is bad (or it's not HTML at all), fall back
+                    // to returning the prettified plain text
+                    if (errorPos != -1) {
+                        cleanHtml = m_currentRoom->prettyPrint(e.plainBody());
+                        static Settings settings;
+                        // A manhole to visualise HTML errors
+                        if (settings.get("Debug/html", false))
+                            cleanHtml +=
+                                QStringLiteral("<br /><font color=\"red\">"
+                                               "At pos %1: %2</font>")
+                                    .arg(QString::number(errorPos), errorString);
+                    }
+                    return cleanHtml;
+                }
+                if (e.hasFileContent()) {
+                    auto fileCaption =
+                        e.content()->fileInfo()->originalName.toHtmlEscaped();
                     if (fileCaption.isEmpty())
                         fileCaption = m_currentRoom->prettyPrint(e.plainBody());
                     return !fileCaption.isEmpty() ? fileCaption : tr("a file");
                 }
                 return m_currentRoom->prettyPrint(e.plainBody());
+                // clang-format off
             }
             , [this] (const RoomMemberEvent& e) {
+                // clang-format on
                 // FIXME: Rewind to the name that was at the time of this event
-                QString subjectName = m_currentRoom->roomMembername(e.userId());
+                const auto subjectName =
+                    m_currentRoom->safeMemberName(e.userId()).toHtmlEscaped();
                 // The below code assumes senderName output in AuthorRole
                 switch( e.membership() )
                 {
                     case MembershipType::Invite:
-                        if (e.repeatsState())
-                            return tr("reinvited %1 to the room").arg(subjectName);
-                        FALLTHROUGH;
-                    case MembershipType::Join:
-                    {
-                        if (e.repeatsState())
-                            return tr("joined the room (repeated)");
-                        if (!e.prevContent() ||
-                                e.membership() != e.prevContent()->membership)
-                        {
-                            return e.membership() == MembershipType::Invite
-                                    ? tr("invited %1 to the room").arg(subjectName)
-                                    : tr("joined the room");
-                        }
+                    case MembershipType::Join: {
                         QString text {};
-                        if (e.isRename())
-                        {
+                        // Part 1: invites and joins
+                        if (e.membership() == MembershipType::Invite)
+                            text = tr("invited %1 to the room")
+                                   .arg(subjectName);
+                        else if (e.changesMembership())
+                            text = tr("joined the room");
+
+                        if (!text.isEmpty()) {
+                            if (e.repeatsState())
+                                text += ' '
+                                    //: State event that doesn't change the state
+                                    % tr("(repeated)");
+                            if (!e.reason().isEmpty())
+                                text += ": " + e.reason().toHtmlEscaped();
+                            return text;
+                        }
+
+                        // Part 2: profile changes of joined members
+                        if (e.isRename()
+                            && Settings().get("UI/show_rename", true)) {
                             if (e.displayName().isEmpty())
                                 text = tr("cleared the display name");
                             else
                                 text = tr("changed the display name to %1")
-                                            .arg(e.displayName());
+                                       .arg(e.displayName().toHtmlEscaped());
                         }
-                        if (e.isAvatarUpdate())
-                        {
+                        if (e.isAvatarUpdate()
+                            && Settings().get("UI/show_avatar_update", true)) {
                             if (!text.isEmpty())
-                                text += " and ";
+                                //: Joiner for member profile updates;
+                                //: mind the leading and trailing spaces!
+                                text += tr(" and ");
                             if (e.avatarUrl().isEmpty())
                                 text += tr("cleared the avatar");
                             else
@@ -479,6 +527,14 @@ QVariant MessageEventModel::data(const QModelIndex& idx, int role) const
                         return text;
                     }
                     case MembershipType::Leave:
+                        if (e.prevContent() &&
+                            e.prevContent()->membership == MembershipType::Invite)
+                        {
+                            return (e.senderId() != e.userId())
+                                    ? tr("withdrew %1's invitation").arg(subjectName)
+                                    : tr("rejected the invitation");
+                        }
+
                         if (e.prevContent() &&
                                 e.prevContent()->membership == MembershipType::Ban)
                         {
@@ -487,11 +543,21 @@ QVariant MessageEventModel::data(const QModelIndex& idx, int role) const
                                     : tr("self-unbanned");
                         }
                         return (e.senderId() != e.userId())
-                                ? tr("has put %1 out of the room").arg(subjectName)
+                                ? e.reason().isEmpty()
+                                  ? tr("kicked %1 from the room")
+                                    .arg(subjectName)
+                                  : tr("kicked %1 from the room: %2")
+                                    .arg(subjectName,
+                                         e.reason().toHtmlEscaped())
                                 : tr("left the room");
                     case MembershipType::Ban:
                         return (e.senderId() != e.userId())
-                                ? tr("banned %1 from the room").arg(subjectName)
+                                ? e.reason().isEmpty()
+                                  ? tr("banned %1 from the room")
+                                    .arg(subjectName)
+                                  : tr("banned %1 from the room: %2")
+                                    .arg(subjectName,
+                                         e.reason().toHtmlEscaped())
                                 : tr("self-banned from the room");
                     case MembershipType::Knock:
                         return tr("knocked");
@@ -499,9 +565,12 @@ QVariant MessageEventModel::data(const QModelIndex& idx, int role) const
                         ;
                 }
                 return tr("made something unknown");
+                // clang-format off
             }
             , [] (const RoomAliasesEvent& e) {
-                return tr("set aliases to: %1").arg(e.aliases().join(", "));
+                return tr("has set room aliases on server %1 to: %2")
+                       .arg(e.stateKey(),
+                            QLocale().createSeparatedList(e.aliases()));
             }
             , [] (const RoomCanonicalAliasEvent& e) {
                 return (e.alias().isEmpty())
@@ -511,12 +580,14 @@ QVariant MessageEventModel::data(const QModelIndex& idx, int role) const
             , [] (const RoomNameEvent& e) {
                 return (e.name().isEmpty())
                         ? tr("cleared the room name")
-                        : tr("set the room name to: %1").arg(e.name());
+                        : tr("set the room name to: %1")
+                          .arg(e.name().toHtmlEscaped());
             }
-            , [] (const RoomTopicEvent& e) {
+            , [this] (const RoomTopicEvent& e) {
                 return (e.topic().isEmpty())
                         ? tr("cleared the topic")
-                        : tr("set the topic to: %1").arg(e.topic());
+                        : tr("set the topic to: %1")
+                          .arg(m_currentRoom->prettyPrint(e.topic()));
             }
             , [] (const RoomAvatarEvent&) {
                 return tr("changed the room avatar");
@@ -524,8 +595,31 @@ QVariant MessageEventModel::data(const QModelIndex& idx, int role) const
             , [] (const EncryptionEvent&) {
                 return tr("activated End-to-End Encryption");
             }
-            , tr("Unknown Event")
+            , [] (const RoomCreateEvent& e) {
+                return (e.isUpgrade()
+                        ? tr("upgraded the room to version %1")
+                        : tr("created the room, version %1")
+                       ).arg(e.version().isEmpty()
+                             ? "1" : e.version().toHtmlEscaped());
+            }
+            , [] (const RoomTombstoneEvent& e) {
+                return tr("upgraded the room: %1")
+                       .arg(e.serverMessage().toHtmlEscaped());
+            }
+            , [] (const StateEventBase& e) {
+                // A small hack for state events from TWIM bot
+                return e.stateKey() == "twim"
+                    ? tr("updated the database", "TWIM bot updated the database")
+                    : e.stateKey().isEmpty()
+                    ? tr("updated %1 state", "%1 - Matrix event type")
+                      .arg(e.matrixType())
+                    : tr("updated %1 state for %2",
+                         "%1 - Matrix event type, %2 - state key")
+                      .arg(e.matrixType(), e.stateKey().toHtmlEscaped());
+            }
+            , tr("Unknown event")
         );
+        // clang-format on
     }
 
     if( role == Qt::ToolTipRole )
@@ -581,10 +675,10 @@ QVariant MessageEventModel::data(const QModelIndex& idx, int role) const
     {
         if (evt.isRedacted())
         {
-            auto reason = evt.redactedBecause()->reason();
+            const auto reason = evt.redactedBecause()->reason();
             return (reason.isEmpty())
                     ? tr("Redacted")
-                    : tr("Redacted: %1").arg(evt.redactedBecause()->reason());
+                    : tr("Redacted: %1").arg(reason.toHtmlEscaped());
         }
 
         if (auto e = eventCast<const RoomMessageEvent>(&evt))
@@ -595,21 +689,28 @@ QVariant MessageEventModel::data(const QModelIndex& idx, int role) const
             return e->hasFileContent()
                     ? QVariant::fromValue(e->content()->originalJson)
                     : QVariant();
-        };
+        }
     }
 
     if( role == HighlightRole )
         return m_currentRoom->isEventHighlighted(&evt);
 
-    if( role == ReadMarkerRole )
-        return evt.id() == lastReadEventId;
-
     if( role == SpecialMarksRole )
     {
-        if (isPending)
-            return pendingIt->deliveryStatus();
+        if (is<RedactionEvent>(evt) || is<ReactionEvent>(evt))
+            return EventStatus::Hidden; // Never show, even pending
 
-        if (is<RedactionEvent>(evt))
+        if (isPending)
+            return !Settings().get<bool>("UI/suppress_local_echo")
+                    ? pendingIt->deliveryStatus() : EventStatus::Hidden;
+
+        // isReplacement?
+        if (auto e = eventCast<const RoomMessageEvent>(&evt))
+            if (!e->replacedEvent().isEmpty())
+                return EventStatus::Hidden;
+
+        if ((is<RoomAliasesEvent>(evt) || is<RoomCanonicalAliasEvent>(evt))
+                && !Settings().value("UI/show_alias_update", true).toBool())
             return EventStatus::Hidden;
 
         auto* memberEvent = timelineIt->viewAs<RoomMemberEvent>();
@@ -618,6 +719,25 @@ QVariant MessageEventModel::data(const QModelIndex& idx, int role) const
             if ((memberEvent->isJoin() || memberEvent->isLeave()) &&
                     !Settings().value("UI/show_joinleave", true).toBool())
                 return EventStatus::Hidden;
+
+            if ((memberEvent->isInvite() || memberEvent->isRejectedInvite())
+                    && !Settings().value("UI/show_invite", true).toBool())
+                return EventStatus::Hidden;
+
+            if ((memberEvent->isBan() || memberEvent->isUnban())
+                    && !Settings().value("UI/show_ban", true).toBool())
+                return EventStatus::Hidden;
+
+            bool hideRename = memberEvent->isRename()
+                && (!memberEvent->isJoin() && !memberEvent->isLeave())
+                && !Settings().value("UI/show_rename", true).toBool();
+            bool hideAvatarUpdate = memberEvent->isAvatarUpdate()
+                && !Settings().value("UI/show_avatar_update", true).toBool();
+            if ((hideRename && hideAvatarUpdate)
+                    || (hideRename && !memberEvent->isAvatarUpdate())
+                    || (hideAvatarUpdate && !memberEvent->isRename())) {
+                return EventStatus::Hidden;
+            }
         }
         if (memberEvent || evt.isRedacted())
         {
@@ -642,7 +762,11 @@ QVariant MessageEventModel::data(const QModelIndex& idx, int role) const
                 !Settings().value("UI/show_noop_events").toBool())
             return EventStatus::Hidden;
 
-        return EventStatus::Normal;
+        if (!evt.isStateEvent() && !is<RoomMessageEvent>(evt)
+                && !Settings().value("UI/show_unknown_events").toBool())
+            return EventStatus::Hidden;
+
+        return evt.isReplaced() ? EventStatus::Replaced : EventStatus::Normal;
     }
 
     if( role == EventIdRole )
@@ -653,13 +777,60 @@ QVariant MessageEventModel::data(const QModelIndex& idx, int role) const
         if (auto e = eventCast<const RoomMessageEvent>(&evt))
             if (e->hasFileContent())
                 return QVariant::fromValue(
-                            m_currentRoom->fileTransferInfo(e->id()));
+                            m_currentRoom->fileTransferInfo(
+                                isPending ? e->transactionId() : e->id()));
     }
 
     if( role == AnnotationRole )
-        if (isPending)
-            return pendingIt->annotation();
+        return isPending ? pendingIt->annotation() : QString();
 
+    if( role == ReactionsRole ) {
+        // Filter reactions out of all annotations and collate them by key
+        struct Reaction {
+            QString key;
+            QStringList authorsList {};
+            bool includesLocalUser = false;
+        };
+        std::vector<Reaction> reactions; // using vector to maintain the order
+        // XXX: Should the list be ordered by the number of reactions instead?
+        const auto& annotations =
+            m_currentRoom->relatedEvents(evt, EventRelation::Annotation());
+        for (const auto& a: annotations)
+            if (const auto e = eventCast<const ReactionEvent>(a)) {
+                auto rIt = std::find_if(reactions.begin(), reactions.end(),
+                                        [&e] (const Reaction& r) {
+                                            return r.key == e->relation().key;
+                                       });
+                if (rIt == reactions.end())
+                    rIt = reactions.insert(reactions.end(), {e->relation().key});
+
+                rIt->authorsList << m_currentRoom->safeMemberName(e->senderId());
+                rIt->includesLocalUser |=
+                        e->senderId() == m_currentRoom->localUser()->id();
+            }
+        // Prepare the QML model data
+        // NB: Strings are NOT HTML-escaped; QML code must take care to use
+        // Text.PlainText format when displaying them
+        QJsonArray qmlReactions;
+        for (auto&& r: reactions) {
+            const auto authorsCount = r.authorsList.size();
+            if (r.authorsList.size() > 7) {
+                //: When the reaction comes from too many members
+                r.authorsList.replace(3, tr("%Ln more member(s)", "",
+                                            r.authorsList.size() - 3));
+                r.authorsList.erase(r.authorsList.begin() + 4,
+                                    r.authorsList.end());
+            }
+            qmlReactions << QJsonObject {
+                    { QStringLiteral("key"), r.key },
+                    { QStringLiteral("authorsCount"), authorsCount },
+                    { QStringLiteral("authors"),
+                            QLocale().createSeparatedList(r.authorsList) },
+                    { QStringLiteral("includesLocalUser"), r.includesLocalUser }
+                };
+        }
+        return qmlReactions;
+    }
 
     if( role == TimeRole || role == SectionRole)
     {
@@ -677,5 +848,10 @@ QVariant MessageEventModel::data(const QModelIndex& idx, int role) const
                                 ? SectionRole : AuthorRole);
         }
 
+    if (role == RefRole)
+        return visit(
+            evt, [](const RoomCreateEvent& e) { return e.predecessor().roomId; },
+            [](const RoomTombstoneEvent& e) { return e.successorRoomId(); });
+
     return {};
 }
