@@ -25,214 +25,15 @@
 #include <connection.h>
 #include <settings.h>
 
+// See a comment in the same place at userlistmodel.cpp
+#include <QtWidgets/QAbstractItemView>
 #include <QtGui/QIcon>
+#include <QtGui/QFontMetrics>
+#include <QtGui/QPalette>
+#include <QtGui/QBrush>
 #include <QtCore/QStringBuilder>
 
-#include <functional>
-
-using namespace std::placeholders;
-
-//const RoomGroups& AbstractRoomOrdering::modelGroups() const
-//{
-//    return model()->m_roomGroups;
-//}
-
-//RoomGroups& AbstractRoomOrdering::modelGroups()
-//{
-//    return model()->m_roomGroups;
-//}
-
-AbstractRoomOrdering::groupLessThan_closure_t
-AbstractRoomOrdering::groupLessThanFactory() const
-{
-    return std::bind(&AbstractRoomOrdering::groupLessThan, this, _1, _2);
-}
-
-AbstractRoomOrdering::roomLessThan_closure_t
-AbstractRoomOrdering::roomLessThanFactory(const QVariant& group) const
-{
-    return std::bind(&AbstractRoomOrdering::roomLessThan, this, group, _1, _2);
-}
-
-void AbstractRoomOrdering::updateGroups(Room* room)
-{
-    model()->updateGroups(room);
-}
-
-template <typename LT, typename VT>
-inline auto findIndex(const QList<LT>& list, const VT& value)
-{
-    // Using std::find() instead of indexOf() so that not found keys were
-    // naturally sorted after found ones.
-    return std::find(list.begin(), list.end(), value) - list.begin();
-}
-
-auto findIndexWithWildcards(const QStringList& list, const QString& value)
-{
-    if (list.empty() || value.isEmpty())
-        return list.size();
-
-    auto i = findIndex(list, value);
-    // Try namespace groupings (".*" in the list), from right to left
-    for (int dotPos = 0;
-         i == list.size() && (dotPos = value.lastIndexOf('.', --dotPos)) != -1;)
-    {
-        i = findIndex(list, value.left(dotPos + 1) + '*');
-    }
-    return i;
-}
-
-const QString OrderByTag::Invite = QStringLiteral("org.qmatrixclient.invite");
-const QString OrderByTag::DirectChat = QStringLiteral("org.qmatrixclient.direct");
-const QString OrderByTag::Untagged = QStringLiteral("org.qmatrixclient.none");
-const QString OrderByTag::Left = QStringLiteral("org.qmatrixclient.left");
-
-QVariant OrderByTag::groupLabel(const RoomGroup& g) const
-{
-    static const auto InvitesLabel = RoomListModel::tr("Invited");
-    static const auto FavouritesLabel = RoomListModel::tr("Favourite");
-    static const auto LowPriorityLabel = RoomListModel::tr("Low priority");
-    static const auto DirectChatsLabel = RoomListModel::tr("People");
-    static const auto UngroupedRoomsLabel = RoomListModel::tr("Ungrouped rooms");
-    static const auto LeftLabel = RoomListModel::tr("Left");
-
-    const auto caption =
-            g.key == Untagged ? UngroupedRoomsLabel :
-            g.key == Invite ? InvitesLabel :
-            g.key == DirectChat ? DirectChatsLabel :
-            g.key == Left ? LeftLabel :
-            g.key == QMatrixClient::FavouriteTag ? FavouritesLabel :
-            g.key == QMatrixClient::LowPriorityTag ? LowPriorityLabel :
-            g.key.toString().startsWith("u.") ? g.key.toString().mid(2) :
-            g.key.toString();
-    return RoomListModel::tr("%1 (%Ln room(s))", "", g.rooms.size()).arg(caption);
-}
-
-bool OrderByTag::groupLessThan(const RoomGroup& g1, const QVariant& g2key) const
-{
-    const auto& lkey = g1.key.toString();
-    const auto& rkey = g2key.toString();
-    // See above
-    auto li = findIndexWithWildcards(tagsOrder, lkey);
-    auto ri = findIndexWithWildcards(tagsOrder, rkey);
-    return li < ri || (li == ri && lkey < rkey);
-}
-
-bool OrderByTag::roomLessThan(const QVariant& groupKey,
-                              const Room* r1, const Room* r2) const
-{
-    if (r1 == r2)
-        return false; // 0. Short-circuit for coinciding room objects
-
-    // 1. Compare tag order values
-    const auto& tag = groupKey.toString();
-    auto o1 = r1->tag(tag).order;
-    auto o2 = r2->tag(tag).order;
-    if (o2.omitted() != o1.omitted())
-        return o2.omitted();
-
-    if (!o1.omitted() && !o2.omitted())
-    {
-        // Compare floats; fallthrough if neither is smaller
-        if (o1.value() < o2.value())
-            return true;
-
-        if (o1.value() > o2.value())
-            return false;
-    }
-
-    // 2. Neither tag order is less than the other; compare room display names
-    if (auto roomCmpRes = r1->displayName().localeAwareCompare(r2->displayName()))
-        return roomCmpRes < 0;
-
-    // 4. Within the same display name, order by room id
-    // (typically the case when both display names are completely empty)
-    if (auto roomIdCmpRes = r1->id().compare(r2->id()))
-        return roomIdCmpRes < 0;
-
-    // 3. Room ids are equal; order by connections (=userids)
-    const auto c1 = r1->connection();
-    const auto c2 = r2->connection();
-    if (c1 != c2)
-    {
-        if (auto usersCmpRes = c1->userId().compare(c2->userId()))
-            return usersCmpRes < 0;
-
-        // 3a. Two logins under the same userid: pervert, but technically correct
-        Q_ASSERT(c1->accessToken() != c2->accessToken());
-        return c1->accessToken() < c2->accessToken();
-    }
-
-    // 5. Assume two incarnations of the room with the different join state
-    // (by design, join states are distinct within one connection+roomid)
-    Q_ASSERT(r1->joinState() != r2->joinState());
-    return r1->joinState() < r2->joinState();
-}
-
-AbstractRoomOrdering::groups_t OrderByTag::roomGroups(const Room* room) const
-{
-    if (room->joinState() == QMatrixClient::JoinState::Invite)
-        return groups_t {{ Invite }};
-    if (room->joinState() == QMatrixClient::JoinState::Leave)
-        return groups_t {{ Left }};
-    auto tags = room->tags().keys();
-    groups_t vl; vl.reserve(tags.size());
-    std::copy(tags.cbegin(), tags.cend(), std::back_inserter(vl));
-    if (room->isDirectChat())
-        vl.push_back(DirectChat);
-    if (vl.empty())
-        vl.push_back(Untagged);
-    return vl;
-}
-
-void OrderByTag::connectSignals(Connection* connection)
-{
-    using DCMap = Connection::DirectChatsMap;
-    QObject::connect( connection, &Connection::directChatsListChanged, model(),
-        [this,connection] (DCMap additions, DCMap removals) {
-            // The same room may show up in removals and in additions if it
-            // moves from one userid to another (pretty weird but encountered
-            // in the wild). Therefore process removals first.
-            for (const auto& rId: removals)
-                if (auto* r = connection->room(rId))
-                    updateGroups(r);
-            for (const auto& rId: additions)
-                if (auto* r = connection->room(rId))
-                    updateGroups(r);
-        });
-}
-
-void OrderByTag::connectSignals(Room* room)
-{
-    QObject::connect(room, &Room::displaynameChanged,
-                     model(), [this,room] { updateGroups(room); });
-    QObject::connect(room, &Room::tagsChanged,
-                     model(), [this,room] { updateGroups(room); });
-
-    QObject::connect(room, &Room::joinStateChanged,
-                     model(), [this,room] { updateGroups(room); });
-}
-
-QStringList OrderByTag::initTagsOrder()
-{
-    using namespace QMatrixClient;
-    static const QStringList DefaultTagsOrder {
-        Invite, FavouriteTag, QStringLiteral("u.*"), DirectChat, Untagged,
-        LowPriorityTag, Left
-    };
-
-    static const auto SettingsKey = QStringLiteral("tags_order");
-    static QMatrixClient::SettingsGroup sg { "UI/RoomsDock" };
-    const auto savedOrder = sg.get<QStringList>(SettingsKey);
-    if (savedOrder.isEmpty())
-    {
-        sg.setValue(SettingsKey, DefaultTagsOrder);
-        return DefaultTagsOrder;
-    }
-    return savedOrder;
-}
-
-RoomListModel::RoomListModel(QObject* parent)
+RoomListModel::RoomListModel(QAbstractItemView* parent)
     : QAbstractItemModel(parent)
 {
     connect(this, &RoomListModel::modelAboutToBeReset,
@@ -241,23 +42,22 @@ RoomListModel::RoomListModel(QObject* parent)
             this, &RoomListModel::restoreCurrentSelection);
 }
 
-void RoomListModel::addConnection(QMatrixClient::Connection* connection)
+void RoomListModel::addConnection(Quotient::Connection* connection)
 {
     Q_ASSERT(connection);
 
-    using namespace QMatrixClient;
+    using namespace Quotient;
     m_connections.emplace_back(connection, this);
-    connect( connection, &Connection::loggedOut,
-             this, [=]{ deleteConnection(connection); } );
-    connect( connection, &Connection::newRoom,
-             this, &RoomListModel::addRoom);
+    connect(connection, &Connection::loggedOut, this,
+            [=] { deleteConnection(connection); });
+    connect(connection, &Connection::newRoom, this, &RoomListModel::addRoom);
     m_roomOrder->connectSignals(connection);
 
-    for (auto* r: connection->roomMap())
+    for (auto* r: connection->allRooms())
         addRoom(r);
 }
 
-void RoomListModel::deleteConnection(QMatrixClient::Connection* connection)
+void RoomListModel::deleteConnection(Quotient::Connection* connection)
 {
     Q_ASSERT(connection);
     const auto connIt =
@@ -269,7 +69,7 @@ void RoomListModel::deleteConnection(QMatrixClient::Connection* connection)
         return;
     }
 
-    for (auto* r: connection->roomMap())
+    for (auto* r: connection->allRooms())
         deleteRoom(r);
     m_connections.erase(connIt);
     connection->disconnect(this);
@@ -285,7 +85,7 @@ void RoomListModel::deleteTag(QModelIndex index)
         qCritical() << "RoomListModel: Invalid tag at position" << index.row();
         return;
     }
-    if (tag.startsWith("org.qmatrixclient."))
+    if (tag.startsWith(RoomGroup::SystemPrefix))
     {
         qWarning() << "RoomListModel: System groups cannot be deleted "
                       "(tried to delete" << tag << "group)";
@@ -296,6 +96,8 @@ void RoomListModel::deleteTag(QModelIndex index)
     for (const auto& c: m_connections)
         for (auto* r: c->roomsWithTag(tag))
             r->removeTag(tag);
+
+    Quotient::SettingsGroup("UI/RoomsDock").remove(tag);
 }
 
 void RoomListModel::visitRoom(const Room& room,
@@ -342,11 +144,14 @@ QModelIndex RoomListModel::indexOf(const QVariant& group) const
 
 QModelIndex RoomListModel::indexOf(const QVariant& group, Room* room) const
 {
-    for (const auto& idx: m_roomIndices.values(room))
+    auto it = m_roomIndices.find(room);
+    if (group.isNull() && it != m_roomIndices.end())
+        return *it;
+    for (;it != m_roomIndices.end() && it.key() == room; ++it)
     {
-        Q_ASSERT(isValidRoomIndex(idx));
-        if (m_roomGroups[idx.parent().row()].key == group)
-            return idx;
+        Q_ASSERT(isValidRoomIndex(*it));
+        if (m_roomGroups[it->parent().row()].key == group)
+            return *it;
     }
     return {};
 }
@@ -371,18 +176,18 @@ QModelIndex RoomListModel::parent(const QModelIndex& child) const
 
 void RoomListModel::addRoom(Room* room)
 {
+    Q_ASSERT(room && !room->id().isEmpty());
     addRoomToGroups(room);
     connectRoomSignals(room);
 }
 
 void RoomListModel::deleteRoom(Room* room)
 {
-    visitRoom(*room, std::bind(&RoomListModel::doRemoveRoom, this, _1));
+    visitRoom(*room, [this] (QModelIndex idx) { doRemoveRoom(idx); });
     room->disconnect(this);
 }
 
-RoomGroups::iterator RoomListModel::tryInsertGroup(const QVariant& key,
-                                                   bool notify)
+RoomGroups::iterator RoomListModel::tryInsertGroup(const QVariant& key)
 {
     Q_ASSERT(!key.toString().isEmpty());
     auto gIt = lowerBoundGroup(key);
@@ -390,41 +195,37 @@ RoomGroups::iterator RoomListModel::tryInsertGroup(const QVariant& key,
     {
         const auto gPos = gIt - m_roomGroups.begin();
         const auto affectedIdxs = preparePersistentIndexChange(gPos, 1);
-        if (notify)
-            beginInsertRows({}, gPos, gPos);
+        beginInsertRows({}, gPos, gPos);
         gIt = m_roomGroups.insert(gIt, {key, {}});
-        if (notify)
-        {
-            endInsertRows();
-            emit groupAdded(gPos);
-        }
+        endInsertRows();
         changePersistentIndexList(affectedIdxs.first, affectedIdxs.second);
+        emit groupAdded(gPos);
     }
+    // Check that the group is healthy
+    Q_ASSERT(gIt->key == key
+             && (gIt->rooms.empty() || !gIt->rooms.front()->id().isEmpty()));
     return gIt;
 }
 
-void RoomListModel::addRoomToGroups(Room* room, bool notify,
-                                    QVariantList groups)
+void RoomListModel::addRoomToGroups(Room* room, QVariantList groups)
 {
     if (groups.empty())
         groups = m_roomOrder->roomGroups(room);
-    for (const auto& g: groups)
+    for (const auto& g: std::as_const(groups))
     {
-        const auto gIt = tryInsertGroup(g, notify);
+        const auto gIt = tryInsertGroup(g);
         const auto rIt = lowerBoundRoom(*gIt, room);
-        if (rIt != gIt->rooms.end() && *rIt == room)
+        if (rIt != gIt->rooms.cend() && *rIt == room)
         {
             qWarning() << "RoomListModel:" << room->objectName()
                        << "is already listed under group" << g.toString();
             continue;
         }
-        const auto rPos = rIt - gIt->rooms.begin();
-        const auto gIdx = index(gIt - m_roomGroups.begin(), 0);
-        if (notify)
-            beginInsertRows(gIdx, rPos, rPos);
+        const auto rPos = int(rIt - gIt->rooms.cbegin());
+        const auto gIdx = index(int(gIt - m_roomGroups.cbegin()), 0);
+        beginInsertRows(gIdx, rPos, rPos);
         gIt->rooms.insert(rIt, room);
-        if (notify)
-            endInsertRows();
+        endInsertRows();
         m_roomIndices.insert(room, index(rPos, 0, gIdx));
         qDebug() << "RoomListModel: Added" << room->objectName()
                  << "to group" << gIt->key.toString();
@@ -445,7 +246,7 @@ void RoomListModel::connectRoomSignals(Room* room)
             this, [this,room] { refresh(room, { Qt::DecorationRole }); });
 }
 
-void RoomListModel::doRemoveRoom(QModelIndex idx)
+void RoomListModel::doRemoveRoom(const QModelIndex &idx)
 {
     if (!isValidRoomIndex(idx))
     {
@@ -454,8 +255,9 @@ void RoomListModel::doRemoveRoom(QModelIndex idx)
         return;
     }
     const auto gPos = idx.parent().row();
-    auto& group = m_roomGroups[gPos];
-    const auto rIt = group.rooms.begin() + idx.row();
+    auto& group = m_roomGroups[gPos]; // clazy:exclude=detaching-member
+    const auto rIt =
+        group.rooms.begin() + idx.row(); // clazy:exclude=detaching-member
     qDebug() << "RoomListModel: Removing room" << (*rIt)->objectName()
              << "from group" << group.key.toString();
     if (m_roomIndices.remove(*rIt, idx) != 1)
@@ -480,17 +282,28 @@ void RoomListModel::doRemoveRoom(QModelIndex idx)
     }
 }
 
-void RoomListModel::doRebuild()
+void RoomListModel::doSetOrder(std::unique_ptr<AbstractRoomOrdering>&& newOrder)
 {
+    beginResetModel();
     m_roomGroups.clear();
     m_roomIndices.clear();
+    if (m_roomOrder)
+        m_roomOrder->deleteLater();
+    m_roomOrder = newOrder.release();
+    endResetModel();
     for (const auto& c: m_connections)
-        for (auto* r: c->roomMap())
-            addRoomToGroups(r, false);
+    {
+        m_roomOrder->connectSignals(c);
+        const auto& allRooms = c->allRooms();
+        for (auto* r: allRooms) {
+            addRoomToGroups(r);
+            m_roomOrder->connectSignals(r);
+        }
+    }
 }
 
 std::pair<QModelIndexList, QModelIndexList>
-RoomListModel::preparePersistentIndexChange(int fromPos, int shiftValue)
+RoomListModel::preparePersistentIndexChange(int fromPos, int shiftValue) const
 {
     QModelIndexList from, to;
     for (auto& pIdx: persistentIndexList())
@@ -519,16 +332,16 @@ int RoomListModel::totalRooms() const
 {
     int result = 0;
     for (const auto& c: m_connections)
-        result += c->roomMap().size();
+        result += c->allRooms().size();
     return result;
 }
 
-bool RoomListModel::isValidGroupIndex(QModelIndex i) const
+bool RoomListModel::isValidGroupIndex(const QModelIndex& i) const
 {
     return i.isValid() && !i.parent().isValid() && i.row() < m_roomGroups.size();
 }
 
-bool RoomListModel::isValidRoomIndex(QModelIndex i) const
+bool RoomListModel::isValidRoomIndex(const QModelIndex& i) const
 {
     return i.isValid() && isValidGroupIndex(i.parent()) &&
             i.row() < m_roomGroups[i.parent().row()].rooms.size();
@@ -539,10 +352,33 @@ QVariant RoomListModel::data(const QModelIndex& index, int role) const
     if (!index.isValid())
         return {};
 
+    const auto* view = static_cast<const QAbstractItemView*>(parent());
     if (isValidGroupIndex(index))
     {
-        if (role == Qt::DisplayRole)
-            return m_roomOrder->groupLabel(m_roomGroups[index.row()]);
+        if (role == Qt::DisplayRole) {
+            int unreadRoomsCount = 0;
+            for (auto &r: m_roomGroups[index.row()].rooms)
+                unreadRoomsCount += r->unreadCount() != -1;
+
+            const auto postfix = unreadRoomsCount
+                ? QStringLiteral(" [%1]").arg(unreadRoomsCount) : QString();
+
+            return m_roomOrder->groupLabel(m_roomGroups[index.row()]).toString()
+                + postfix;
+        }
+
+        // It would be more proper to do it in RoomListItemDelegate
+        // (see roomlistdock.cpp) but I (@kitsune) couldn't find a working way.
+        if (role == Qt::BackgroundRole)
+            return view->palette().brush(QPalette::Active, QPalette::Button);
+
+        if (role == HighlightCountRole) {
+            int highlightCount = 0;
+            for (auto &r: m_roomGroups[index.row()].rooms)
+                highlightCount += r->highlightCount();
+
+            return highlightCount;
+        }
 
         return {};
     }
@@ -555,76 +391,120 @@ QVariant RoomListModel::data(const QModelIndex& index, int role) const
 //    if (index.column() == 2)
 //        return room->lastAttended();
 
-    using QMatrixClient::JoinState;
+    static const auto RoomNameTemplate = tr("%1 (as %2)", "%Room (as %user)");
+    auto disambiguatedName = room->displayName();
+    if (role == Qt::DisplayRole || role == Qt::ToolTipRole)
+        for (const auto& c: m_connections)
+            if (c != room->connection()
+                && c->room(room->id(), room->joinState()))
+                disambiguatedName =
+                    RoomNameTemplate.arg(room->displayName(),
+                                         room->localUser()->id());
+
+    using Quotient::JoinState;
     switch (role)
     {
-        case Qt::DisplayRole:
-        {
-            const auto unreadCount = room->unreadCount();
-            const auto postfix = unreadCount == -1 ? QString() :
-                room->readMarker() != room->timelineEdge()
-                    ? QStringLiteral(" [%1]").arg(unreadCount)
-                    : QStringLiteral(" [%1+]").arg(unreadCount);
-            for (const auto& c: m_connections)
-            {
-                if (c == room->connection())
-                    continue;
-                if (c->room(room->id(), room->joinState()))
-                    return tr("%1 (as %2)").arg(room->displayName(),
-                                                room->connection()->userId())
-                           + postfix;
+        case Qt::DisplayRole: {
+            static Quotient::Settings settings;
+            auto unreadCount = room->unreadCount();
+            auto notifCount = room->notificationCount();
+            auto unreadTilReadReceipt = unreadCount - notifCount;
+            QString value =
+                (room->isUnstable() ? "(!)" : "") % disambiguatedName;
+            if (unreadCount >= 0)
+                value += " ["
+                    % (unreadTilReadReceipt > 0
+                           ? QLocale().toString(unreadTilReadReceipt)
+                           : QString())
+                    % (room->readMarker() == room->historyEdge() ? "?" : "")
+                    % (notifCount > 0 ? QStringLiteral("+%L1").arg(notifCount)
+                                      : QString())
+                    % ']';
+            if (settings.get("Debug/read_receipts", false)) {
+                auto localReadReceipt = room->readMarker(room->localUser());
+                value += " <"
+                    % QString::number(room->historyEdge() - localReadReceipt)
+                    % '|'
+                    % QString::number(room->syncEdge() - localReadReceipt.base())
+                    % '>';
             }
-            return room->displayName() + postfix;
+            return value;
         }
         case Qt::DecorationRole:
         {
-            auto avatar = room->avatar(16, 16);
-            if (!avatar.isNull())
-                return avatar;
-            switch( room->joinState() )
-            {
-                case JoinState::Join:
-                    return QIcon(":/irc-channel-joined.svg");
-                case JoinState::Invite:
-                    return QIcon(":/irc-channel-invited.svg");
-                case JoinState::Leave:
-                    return QIcon(":/irc-channel-parted.svg");
+            const auto dpi = view->devicePixelRatioF();
+            if (auto avatar = room->avatar(int(view->iconSize().height() * dpi));
+                !avatar.isNull()) {
+                avatar.setDevicePixelRatio(dpi);
+                return QIcon(QPixmap::fromImage(avatar));
             }
+            switch (room->joinState()) {
+            case JoinState::Join:
+                return QIcon::fromTheme("user-available",
+                                        QIcon(":/irc-channel-joined"));
+
+            case JoinState::Invite:
+                return QIcon::fromTheme("contact-new",
+                                        QIcon(":/irc-channel-invited"));
+            case JoinState::Leave:
+                return QIcon::fromTheme("user-offline",
+                                        QIcon(":/irc-channel-parted"));
+            default:
+                Q_ASSERT(false); // Unknown JoinState?
+            }
+            return {}; // Shouldn't reach here
         }
         case Qt::ToolTipRole:
         {
-            auto result = QStringLiteral("<b>%1</b><br>").arg(room->displayName());
-            result += tr("Main alias: %1<br>").arg(room->canonicalAlias());
-            result += tr("Members: %1<br>").arg(room->memberCount());
+            QString result =
+                "<b>" % disambiguatedName.toHtmlEscaped() % "</b><br>"
+                % tr("Main alias: %1").arg(room->canonicalAlias().toHtmlEscaped())
+                % "<br>" %
+                //: The number of joined members
+                tr("Joined: %L1").arg(room->joinedCount());
+            if (room->invitedCount() > 0)
+                result += //: The number of invited users
+                    "<br>" % tr("Invited: %L1").arg(room->invitedCount());
 
-            auto directChatUsers = room->directChatUsers();
-            if (!directChatUsers.isEmpty())
-            {
+            const auto directChatUsers = room->directChatUsers();
+            if (!directChatUsers.isEmpty()) {
                 QStringList userNames;
+                userNames.reserve(directChatUsers.size());
                 for (auto* user: directChatUsers)
-                    userNames.push_back(user->displayname(room));
-                result += tr("Direct chat with %1<br>")
-                            .arg(userNames.join(','));
+                    userNames.push_back(user->displayname(room).toHtmlEscaped());
+                result += "<br>"
+                          % tr("Direct chat with %1")
+                                .arg(QLocale().createSeparatedList(userNames));
             }
 
             if (room->usesEncryption())
-                result += tr("The room enforces encryption<br>");
+                result += "<br>" % tr("The room enforces encryption");
 
-            auto unreadCount = room->unreadCount();
-            if (unreadCount >= 0)
-            {
-                const auto unreadLine =
-                    room->readMarker() == room->timelineEdge()
-                        ? tr("Unread messages: %1+<br>")
-                        : tr("Unread messages: %1<br>");
-                result += unreadLine.arg(unreadCount);
+            if (room->isUnstable()) {
+                result += "<br>(!) " % tr("This room's version is unstable!");
+                if (room->canSwitchVersions())
+                    result += ' ' % tr("Consider upgrading to a stable version"
+                                       " (use room settings for that)");
             }
 
-            auto hlCount = room->highlightCount();
-            if (hlCount > 0)
-                result += tr("Unread highlights: %1<br>").arg(hlCount);
+            if (const auto unreadCount = room->unreadCount(); unreadCount >= 0) {
+                // TODO for 0.0.96: tr("Messages after fully read marker: %L1")
+                result += "<br/>" % tr("Unread messages: %L1")
+                                    .arg(unreadCount);
+                if (room->readMarker() == room->historyEdge())
+                    result += ' ' % /*: Unread messages */ tr("(maybe more)");
+            }
 
-            result += tr("ID: %1<br>").arg(room->id());
+            // TODO for 0.0.96: tr("Unread notifications since read receipt: %L1")
+            if (const auto nfCount = room->notificationCount(); nfCount > 0)
+                result += "<br>" % tr("Unread notifications: %L1").arg(nfCount);
+
+            if (const auto hlCount = room->highlightCount(); hlCount > 0)
+                result += "<br>" % tr("Unread highlights: %L1").arg(hlCount);
+
+            // Room ids are pretty safe from rogue HTML; escape it just in case
+            result +=
+                "<br>" % tr("ID: %1").arg(room->id().toHtmlEscaped()) % "<br>";
             switch (room->joinState())
             {
                 case JoinState::Join:
@@ -639,11 +519,13 @@ QVariant RoomListModel::data(const QModelIndex& index, int role) const
             return result;
         }
         case HasUnreadRole:
-            return room->hasUnreadMessages();
+            return room->unreadCount() > 0;
         case HighlightCountRole:
             return room->highlightCount();
         case JoinStateRole:
-            return toCString(room->joinState()); // FIXME: better make the enum QVariant-convertible (only possible from Qt 5.8, see Q_ENUM_NS)
+            if (!room->successorId().isEmpty())
+                return QStringLiteral("upgraded");
+            return toCString(room->joinState()); // TODO: drop toCString once on lib 0.7
         case ObjectRole:
             return QVariant::fromValue(room);
         default:
@@ -658,10 +540,9 @@ int RoomListModel::columnCount(const QModelIndex&) const
 
 void RoomListModel::updateGroups(Room* room)
 {
-    const auto oldRoomIndices = m_roomIndices.values(room);
-    Q_ASSERT(!oldRoomIndices.empty()); // The room should have been somewhere
-
     auto groups = m_roomOrder->roomGroups(room);
+
+    const auto oldRoomIndices = m_roomIndices.values(room);
     for (const auto& oldIndex: oldRoomIndices)
     {
         Q_ASSERT(isValidRoomIndex(oldIndex));
@@ -687,7 +568,7 @@ void RoomListModel::updateGroups(Room* room)
             doRemoveRoom(oldIndex); // May invalidate `group` and `gIdx`
     }
     if (!groups.empty())
-        addRoomToGroups(room, true, groups); // Groups the room wasn't before
+        addRoomToGroups(room, groups); // Groups the room wasn't before
     qDebug() << "RoomListModel: groups for" << room->objectName() << "updated";
 }
 
@@ -696,6 +577,8 @@ void RoomListModel::refresh(Room* room, const QVector<int>& roles)
     // The problem here is that the change might cause the room to change
     // its groups. Assume for now that such changes are processed elsewhere
     // where details about the change are available (e.g. in tagsChanged).
-    visitRoom(*room,
-        [this,&roles] (QModelIndex idx) { emit dataChanged(idx, idx, roles); });
+    visitRoom(*room, [this,&roles] (const QModelIndex &idx) {
+        emit dataChanged(idx, idx, roles);
+        emit dataChanged(idx.parent(), idx.parent(), roles);
+    });
 }
