@@ -21,7 +21,9 @@
 
 #include <QtCore/QAbstractListModel>
 
-namespace QMatrixClient
+class QAbstractItemView;
+
+namespace Quotient
 {
     class Connection;
     class Room;
@@ -32,25 +34,31 @@ class UserListModel: public QAbstractListModel
 {
         Q_OBJECT
     public:
-        using User = QMatrixClient::User;
+        using User = Quotient::User;
 
-        UserListModel(QObject* parent = nullptr);
+        UserListModel(QAbstractItemView* parent);
         virtual ~UserListModel();
 
-        void setRoom(QMatrixClient::Room* room);
+        void setRoom(Quotient::Room* room);
         User* userAt(QModelIndex index);
 
         QVariant data(const QModelIndex& index, int role = Qt::DisplayRole) const override;
         int rowCount(const QModelIndex& parent=QModelIndex()) const override;
+    
+    signals:
+        void membersChanged(); //< Reflection of Room::memberListChanged
+
+    public slots:
+        void filter(const QString& filterString);
 
     private slots:
         void userAdded(User* user);
         void userRemoved(User* user);
         void refresh(User* user, QVector<int> roles = {});
-        void avatarChanged(User* user, const QMatrixClient::Room* context);
+        void avatarChanged(User* user, const Quotient::Room* context);
 
     private:
-        QMatrixClient::Room* m_currentRoom;
+        Quotient::Room* m_currentRoom;
         QList<User*> m_users;
 
         int findUserPos(User* user) const;
