@@ -20,11 +20,14 @@
 
 #include "dialog.h"
 
-namespace QMatrixClient {
+namespace Quotient {
     class Connection;
 }
 
+class MainWindow;
 class QuaternionRoom;
+class AccountRegistry;
+class AccountSelector;
 
 class QComboBox;
 class QLineEdit;
@@ -33,23 +36,22 @@ class QCheckBox;
 class QPushButton;
 class QListWidget;
 class QFormLayout;
+class QStandardItemModel;
 
 class RoomDialogBase : public Dialog
 {
         Q_OBJECT
     protected:
-        using connections_t = QVector<QMatrixClient::Connection*>;
+        using Connection = Quotient::Connection;
 
         RoomDialogBase(const QString& title, const QString& applyButtonText,
-            QuaternionRoom* r, QWidget* parent, const connections_t& cs = {},
+            QuaternionRoom* r, QWidget* parent,
             QDialogButtonBox::StandardButtons extraButtons = QDialogButtonBox::Reset);
 
     protected:
-        const connections_t connections;
         QuaternionRoom* room;
 
         QLabel* avatar;
-        QComboBox* account;
         QLineEdit* roomName;
         QLabel* aliasServer;
         QLineEdit* alias;
@@ -57,20 +59,29 @@ class RoomDialogBase : public Dialog
         QString previousTopic;
         QCheckBox* publishRoom;
         QCheckBox* guestCanJoin;
-        QFormLayout* formLayout;
+        QFormLayout* mainFormLayout;
+        QFormLayout* essentialsLayout = nullptr;
+
+        QComboBox* addVersionSelector(QLayout* layout);
+        void refillVersionSelector(QComboBox* selector, Connection* account);
+        void addEssentials(QWidget* accountControl, QLayout* versionBox);
+        bool checkRoomVersion(QString version, Connection* account);
 };
 
 class RoomSettingsDialog : public RoomDialogBase
 {
         Q_OBJECT
     public:
-        RoomSettingsDialog(QuaternionRoom* room, QWidget* parent = nullptr);
+        RoomSettingsDialog(QuaternionRoom* room, MainWindow* parent = nullptr);
 
     private slots:
         void load() override;
+        bool validate() override;
         void apply() override;
 
     private:
+        QLabel* account;
+        QLabel* version;
         QListWidget* tagsList;
         bool userChangedAvatar = false;
 };
@@ -79,7 +90,7 @@ class CreateRoomDialog : public RoomDialogBase
 {
         Q_OBJECT
     public:
-        CreateRoomDialog(const connections_t& connections,
+        CreateRoomDialog(const AccountRegistry* accounts,
                          QWidget* parent = nullptr);
 
     public slots:
@@ -87,11 +98,16 @@ class CreateRoomDialog : public RoomDialogBase
 
     private slots:
         void load() override;
+        bool validate() override;
         void apply() override;
         void accountSwitched();
 
     private:
+        AccountSelector* accountChooser;
+        QComboBox* version;
         QComboBox* nextInvitee;
         QPushButton* inviteButton;
         QListWidget* invitees;
+
+        QHash<Connection*, QStandardItemModel*> userLists;
 };
